/**************************************
  Mass storage implementation for PIC
        Based on the USB spec.
    Tested under Linux and Windows
     and verified using wireshark
 -------------------------------------
 Copyright David Guillen Fandos 2011
 -------------------------------------
       Use at your own risk!!
***************************************/

#include <pic18fregs.h>
#include "msd.h"
#include "usb.h"
#include "hid_macros.h"
#include "sd.h"

#define NUM_BLOCKS_CAP    236543
#define BLOCK_SIZE        512

unsigned char MSDState;
volatile unsigned char no_media;

unsigned char * data_to_send;
unsigned long data_size;
unsigned long data_transfer_length;
unsigned long num_blocks_req = 0;
unsigned long lba = 0;
unsigned long residue_left;

#pragma udata usbram5  ep1_out_buffer ep1_in_buffer
volatile unsigned char ep1_in_buffer[MSD_DATA_SIZE];
volatile unsigned char ep1_out_buffer[MSD_DATA_SIZE];

const InquiryStruct inq_resp = {
	0x05,0x80,0x04,0x02,
	0x20, {0x00, 0x00, 0x00},
	{'d','a','v','i','d','g','f',' '},
	{'M','a','s','s',' ','S','t','o','r','a','g','e',' ',' ',' ',' '},
	{'0','0','0','1'}
};
const MSDPages dev_pages = {
	{ 32,0,{0,0,0,0,0} },
	{
	0x05,0x1E,
	0x00F0, 0xFF, 0x20,
	BIG_ENDIAN(0x0200),
	0x0EF5,
	{0,0,0,0,0,0,0,0,0},
	0,0,
	{0,0,0,0,0,0,0},
	0,
	{0,0}
	}
};

const StdSenseData illegal_req = {0x70,0x00,0x05,0x0,0x0a,0x0,0x24,0x00,0x0};
const StdSenseData no_medium = { 0xF0, 0x00, 0x06, 0x00, 0x0a, 0x0, 0x3a, 0x00, 0x0 };  // Last long may be 02 54 00 02

const unsigned long capacity[2] = {BIG_ENDIAN4(NUM_BLOCKS_CAP),BIG_ENDIAN4(BLOCK_SIZE)};
const unsigned long format_capacity[2] = {BIG_ENDIAN4(NUM_BLOCKS_CAP),(0x2<<24)|BIG_ENDIAN4(BLOCK_SIZE)};
const unsigned long lun_list[4] = {BIG_ENDIAN4(8),0x0, 0x0,0x0};

//#define CD_PROFILE     0x0008
#define CD_PROFILE     0x0000
#define CD_CONFIG_SIZE 44   // 48 bytes minus 4 header bytes
const unsigned char cd_config[] = { 0x00,0x00, HI_B(CD_CONFIG_SIZE), LO_B(CD_CONFIG_SIZE), 0,0,
			   HI_B(CD_PROFILE),LO_B(CD_PROFILE),

			   0x00,0x00,0x03,0x04,					// Feature 0: List of profles
			   HI_B(CD_PROFILE),LO_B(CD_PROFILE),0x01,0x00,		// Just one profile: CD profile

			   0x00,0x01,0x03,0x04,					// Feature 1: Core feature
			   0x00,0x00,0x00,0x01,					// Using SCSI physical interface

			   0x00,0x03,0x03,0x04,					// Feature 3: Removable medium
			   0x00,0x00,0x00,0x00,					// Removable, no prevent, no eject

			   0x00,0x10,0x03,0x08,					// Feature 10h: Random access
			   0x00,0x00,HI_B(BLOCK_SIZE),LO_B(BLOCK_SIZE),		// logical block size
			   0x00,0x01,0x00,0x00,					// No err recovery, blocking: 1

			   0x00,0x1E,0x01,0x00					// Feature 1Eh: Read, Removable and present
};

const unsigned char cd_toc[] = {
	0x00,0x0a,
	0x01,0x01,
	0x00,0x14,0x01,0x00,   // Main track 00
	0x00,0x00,0x00,0x00,   // Start LBA at 0
	0x00,0x14,0xAA,0x00,   // Lead out track, at the end of the disc?
	HI3_B(NUM_BLOCKS_CAP),HI2_B(NUM_BLOCKS_CAP),HI_B(NUM_BLOCKS_CAP),LO_B(NUM_BLOCKS_CAP)
};

// CWS block contains magic number, tag, extra size and status code
CSWStruct csw_response = {
	{0x55, 0x53, 0x42, 0x53},
	{0,0,0,0},
	0,
	0
};

// Initialize the MSD subsystem (after a SET CONFIG request)
void MSDInitEndpoint(void) {
	int i;
	for (i = 0; i < MSD_DATA_SIZE; i++) {
		ep1_in_buffer[i] = 0;
		ep1_out_buffer[i] = 0;
	}

	MSDState = MSD_STATE_WAIT;
	// Turn on both in and out for this endpoint
	UEP1 = 0x1E;

	BD_PREPARE_OUT(1,MSD_DATA_SIZE,ep1_out_buffer,TDATA0)

	endps[1].in.ADDR = PTR16(ep1_in_buffer);
	endps[1].in.Stat = DTS;
	endps[1].in.Cnt  = MSD_DATA_SIZE;
}

// Process MSD specific requests
char ProcessMSDRequest(void) {
	unsigned char req,reqType;

	// Only process packets directed to 0 interface
	if((SetupPacket.bmRequestType & 0x1F) != RECIP_INTERFACE || (SetupPacket.wIndex0 != 0)) return 0;

	req = SetupPacket.bRequest;
	reqType = (SetupPacket.bmRequestType >> 5) & 0x3;

	if (reqType == REQ_TYPE_CLASS) {
		switch (req) {
		case MSD_RESET:
			MSDState = MSD_STATE_WAIT;
			return 1;
		case GET_MAX_LUN:
			controlTransferBuffer[0] = 0;
			outPtr = controlTransferBuffer;
			wCount = 1;
			return 1;
		default:
			return 0;
		};
	}
}

// An output transfer occured, check if we need to do some work
void MSDOutNotify(void) {}

extern unsigned char block_buffer[512];

unsigned char SCSICommand(unsigned char * comm, unsigned char length) {
	switch (comm[0]) {
	case MSD_COMM_INQUIRY:
		data_to_send = (char*)inq_resp;
		data_size = sizeof(inq_resp);
		return MSD_STATE_OUT;
	case MSD_COMM_READ_CAPACITY:
		data_to_send = (char*)capacity;
		data_size = sizeof(capacity);
		return MSD_STATE_OUT;
	case MSD_COMM_READ_FORMAT_CAPACITY:
		data_to_send = (char*)format_capacity;
		data_size = sizeof(format_capacity);
		return MSD_STATE_OUT;
	case MSD_COMM_TEST_UNIT_READY:
		if (no_media) csw_response.status = 1;
		data_size = 0;
		return MSD_STATE_OUT;
	case MSD_COMM_MODE_SENSE:
	case MSD_COMM_MODE_SENSE_5a:
		data_size = sizeof(dev_pages);
		data_to_send = (char*)dev_pages;
		return MSD_STATE_OUT;
	case MSD_COMM_PREVENT_ALLOW_MEDIUM_REMOVAL:
		if ((comm[4]&0x1) == 1) csw_response.status = 1;
		data_size = 0;
		return MSD_STATE_OUT;
	case MSD_COMM_REQUEST_SENSE:
		if (no_media) {
			data_size = sizeof(no_medium);
			data_to_send = (char*)no_medium;
		}else{
			data_size = sizeof(illegal_req);
			data_to_send = (char*)illegal_req;
		}
		return MSD_STATE_OUT;
	case MSD_COMM_READ_10: {
		lba = 	((unsigned long)comm[5]) | 
			(((unsigned long)comm[4])<<8) | 
			(((unsigned long)comm[3])<<16) | 
			(((unsigned long)comm[2])<<24);
		num_blocks_req = ((unsigned long)comm[8]) | (((unsigned long)comm[7])<<8);

		data_size = (((unsigned long)num_blocks_req)<<9);
		} return MSD_STATE_DATA_OUT;
	case MSD_COMM_WRITE_10: {
		lba = 	((unsigned long)comm[5]) | 
			(((unsigned long)comm[4])<<8) | 
			(((unsigned long)comm[3])<<16) | 
			(((unsigned long)comm[2])<<24);
		num_blocks_req = ((unsigned long)comm[8]) | (((unsigned long)comm[7])<<8);

		data_size = (((unsigned long)num_blocks_req)<<9);
		data_to_send = (char*)block_buffer;
		BD_PREPARE_OUT_TOGGLE(1,MSD_DATA_SIZE,data_to_send)

		} return MSD_STATE_DATA_IN;
	case MSD_COMM_LIST_LUNS:
		data_size = sizeof(lun_list);
		data_to_send = (char*)lun_list;
		return MSD_STATE_OUT;
	case MSD_COMM_GET_CONFIGURATION:
		data_size = sizeof(cd_config);
		data_to_send = cd_config;
		return MSD_STATE_OUT;
	case MSD_COMM_GET_TOC:
		data_size = sizeof(cd_toc);
		data_to_send = (char*)cd_toc;
		return MSD_STATE_OUT;
	case MSD_COMM_READ_DISC_INFO:
		data_size = 2;
		block_buffer[0] = 0;
		block_buffer[1] = 1;
		data_to_send = block_buffer;
		return MSD_STATE_OUT;
	case MSD_COMM_REPORT_KEY:
		csw_response.status = 1;
		data_size = 8;
		data_to_send = block_buffer;
		return MSD_STATE_OUT;
	case MSD_COMM_VERIFY:
	case MSD_COMM_STOP_START:
	default:
		csw_response.status = 1;
		data_size = 0;
		return MSD_STATE_OUT;
	}
	return MSD_STATE_WAIT; // Invalid (do nothing)
}

unsigned char CBWProcess() {
	unsigned char cbw_len;
	CBWStruct * cbw = (CBWStruct *)ep1_out_buffer;

	// Check CBW Signature
/*	if (	ep1_out_buffer[0] != 0x55 ||
		ep1_out_buffer[1] != 0x53 ||
		ep1_out_buffer[2] != 0x42 ||
		ep1_out_buffer[3] != 0x43) return MSD_STATE_WAIT;*/
	if (	cbw->sig[0] != 0x55 || cbw->sig[1] != 0x53 ||
		cbw->sig[2] != 0x42 || cbw->sig[3] != 0x43)
	return MSD_STATE_WAIT;

	CopyMemory(csw_response.tag,&ep1_out_buffer[4],4);
	data_transfer_length = 	(( unsigned long)ep1_out_buffer[8]) | 
				(((unsigned long)ep1_out_buffer[9])<<8) | 
				(((unsigned long)ep1_out_buffer[10])<<16) | 
				(((unsigned long)ep1_out_buffer[11])<<24);

	cbw_len = ep1_out_buffer[14]&0x1F;

	return SCSICommand(&ep1_out_buffer[15],cbw_len);
}

void ClassTransfer() {
	if (MSDState == MSD_STATE_WAIT) {
		if (endps[1].out.Stat & UOWN) return;
		csw_response.status = 0;
		MSDState = CBWProcess(); // Execute command received
		if (data_size > data_transfer_length) data_size = data_transfer_length;
		csw_response.residue = data_transfer_length - data_size;  // Calculate residue data
		residue_left = csw_response.residue;
	}

	// Processes basic data in out. The data is a little buffer (capacity, additiona info, etc.)
	if (MSDState == MSD_STATE_IN) {
		
	}else if (MSDState == MSD_STATE_OUT) {
		if (endps[1].in.Stat & UOWN) return;
		if (data_size > 0) { // Have data pending to send
			unsigned long sts = MSD_DATA_SIZE;
			if (data_size < sts) sts = data_size;

			ep_send(1,data_to_send,sts);
			data_size -= sts;
			data_to_send += sts;
		}else{
			// Send an CSW unless we need to send a data block
			if (num_blocks_req == 0) {
				MSDState = MSD_STATE_CSW;
			}else{
				
			}
		}
	}

	// Processes DATA STREAM OUT, which handles media read/write also
	if (MSDState == MSD_STATE_DATA_IN) {
		if (endps[1].out.Stat & UOWN) return;  // No data received
		if (data_size != 0) {
			if (data_size < endps[1].out.Cnt) data_size = 0;
			else data_size -= endps[1].out.Cnt;
			data_to_send += endps[1].out.Cnt;

			// Completed a block or end of transmission
			if ((data_size&0x1FF) == 0) {
				PORTA = 1;
				WriteSDBlock(block_buffer,lba);
				lba++;
				num_blocks_req--;
				data_to_send = (char*)block_buffer;
			}
			BD_PREPARE_OUT_TOGGLE(1,MSD_DATA_SIZE,data_to_send)
		}
		if (data_size == 0) {
			MSDState = MSD_STATE_CSW;
			// Need to save back the last block received
			BD_PREPARE_OUT_TOGGLE(1,MSD_DATA_SIZE,data_to_send)
		}
	}else if (MSDState == MSD_STATE_DATA_OUT) {
		if (endps[1].in.Stat & UOWN) return;  // Can't send more data
		if (data_size == 0) {
			MSDState = MSD_STATE_CSW;
		}else{
			unsigned long sts = MSD_DATA_SIZE;
			if ((data_size&0x1FF) == 0) {
				// Need to load next block
				ReadSDBlock(block_buffer,lba);
				lba++;
				num_blocks_req--;
				data_to_send = (char*)block_buffer;
			}
			if (data_size < sts) sts = data_size;
			ep_send(1,data_to_send,sts);
			data_size -= sts;
			data_to_send += sts;
		}
	}

	if (MSDState == MSD_STATE_CSW) {  // Send end of transaction
		// Send padding data if needed
		//residue_left = 0;
		//if (residue_left > 0) {
		//	unsigned long sts = MSD_DATA_SIZE;
		//	if (residue_left < MSD_DATA_SIZE) sts = residue_left;
		//	ep_send(1,ep1_out_buffer,sts);

		//	residue_left -= sts;
		//	if (residue_left == 0) 
		//		endps[1].in.Stat = UOWN | BSTALL;
		//}else{
			ep_send(1,(char*)&csw_response,sizeof(CSWStruct));
			MSDState = MSD_STATE_WAIT;
			BD_PREPARE_OUT_TOGGLE(1,MSD_DATA_SIZE,ep1_out_buffer)
		//}
	}
}

