
#ifndef USB_H
#define USB_H

#define LO_B(x) (x&0xFF)
#define HI_B(x) ((x>>8)&0xFF)
#define HI2_B(x) ((x>>16)&0xFF)
#define HI3_B(x) ((x>>24)&0xFF)
#define PTR16(x) ((unsigned int)(((unsigned long)x) & 0xFFFF))

// USB constants

#define GET_STATUS         0
#define CLEAR_FEATURE      1
#define SET_FEATURE        3
#define SET_ADDRESS        5
#define GET_DESCRIPTOR     6
#define SET_DESCRIPTOR     7
#define GET_CONFIGURATION  8
#define SET_CONFIGURATION  9
#define GET_INTERFACE     10
#define SET_INTERFACE     11
#define SYNCH_FRAME       12

// Descriptor Types
#define DEVICE_DESCRIPTOR        0x01
#define CONFIGURATION_DESCRIPTOR 0x02
#define STRING_DESCRIPTOR        0x03
#define INTERFACE_DESCRIPTOR     0x04
#define ENDPOINT_DESCRIPTOR      0x05

// Class Descriptor Types
#define HID_DESCRIPTOR      0x21
#define REPORT_DESCRIPTOR   0x22
#define PHYSICAL_DESCRIPTOR 0x23

// HID Class specific requests
#define GET_REPORT      0x01
#define GET_IDLE        0x02
#define GET_PROTOCOL    0x03
#define SET_REPORT      0x09
#define SET_IDLE        0x0A
#define SET_PROTOCOL    0x0B

// Standard Feature Selectors
#define DEVICE_REMOTE_WAKEUP    0x01
#define ENDPOINT_HALT           0x00

#define UOWN   0x80 // USB Own Bit
#define DTS    0x40 // Data Toggle Synchronization Bit
#define KEN    0x20 // BD Keep Enable Bit
#define INCDIS 0x10 // Address Increment Disable Bit
#define DTSEN  0x08 // Data Toggle Synchronization Enable Bit
#define BSTALL 0x04 // Buffer Stall Enable Bit
#define BC9    0x02 // Byte count bit 9
#define BC8    0x01 // Byte count bit 8

#define TDATA0 0x00
#define TDATA1 0x40

#define USB_DETACHED     0
#define USB_ATTACHED     1
#define USB_POWERED      2
#define USB_DEFAULT      3
#define USB_ADDRESS      4
#define USB_CONFIGURED   5

#define RECIP_DEVICE     0
#define RECIP_INTERFACE  1
#define RECIP_ENDPOINT   2

#define REQ_TYPE_STD     0
#define REQ_TYPE_CLASS   1
#define REQ_TYPE_VENDOR  2

// String def Macro
#define CONCAT(...) __VA_ARGS__
#define USB_STRING(name,str) \
	const unsigned short name[] = { \
	    sizeof(name) | (STRING_DESCRIPTOR<<8), str \
	};


// Endpoint descriptors
typedef struct _BDT {
	unsigned char Stat;
	unsigned char Cnt;
	unsigned short ADDR;
} BDT;

typedef struct _ENDP {
	BDT out,in;
} ENDP;

// USB state variables
extern unsigned char deviceState;
extern unsigned char selfPowered;
extern unsigned char remoteWakeup;
extern unsigned char devConfigured;

extern volatile ENDP at 0x0400 endps[16];

// Size of the buffer for endpoint 0
#define E0SZ 32

typedef struct _setupPacketStruct {
	unsigned char bmRequestType;
	unsigned char bRequest;
	unsigned char wValue0;
	unsigned char wValue1;
	unsigned char wIndex0;
	unsigned char wIndex1;
	unsigned short wLength;
	unsigned char extra[E0SZ-8];
} setupPacketStruct;

extern volatile setupPacketStruct SetupPacket;
extern volatile char controlTransferBuffer[E0SZ];

// Buffer to pass in/out information
extern unsigned char *outPtr;
extern unsigned char *inPtr;
extern short wCount;

// USB Functions
void EnableUSBModule(void);
void ProcessUSBTransactions(void);

// Functions for reading/writing the HID interrupt endpoint
unsigned char ep_send(unsigned char ep, unsigned char * buffer, unsigned short len);
unsigned char ep_recv(unsigned char ep, unsigned char * buffer, unsigned short len);

void ProcessTransfer(unsigned char endp);
void InitializeUSB();

#define DEVICE_DESCRIPTOR_SIZE 0x12
#define CONFIG_HEADER_SIZE  9
#define MSD_DESCRIPTOR_SIZE 23
#define MSD_HEADER_SIZE     9

typedef struct _configStruct {
	unsigned char configHeader[CONFIG_HEADER_SIZE];
	unsigned char HIDDescriptor[MSD_DESCRIPTOR_SIZE];
} ConfigStruct;
extern code ConfigStruct configDescriptor;

#define BD_PREPARE_OUT(epn,epsize,buffer,dts) \
	endps[epn].out.Cnt = LO_B(epsize); \
	endps[epn].out.ADDR = PTR16(buffer); \
	endps[epn].out.Stat = UOWN | DTSEN | dts | (HI_B(epsize)&0x3);

#define BD_PREPARE_OUT_TOGGLE(epn,epsize,buffer) \
	endps[epn].out.Cnt = LO_B(epsize); \
	endps[epn].out.ADDR = PTR16(buffer); \
	if (endps[epn].out.Stat & DTS) endps[epn].out.Stat = UOWN | DTSEN | (HI_B(epsize)&0x3); \
	else endps[epn].out.Stat = UOWN | DTSEN | DTS | (HI_B(epsize)&0x3);

void CopyMemory(char * dst, char * src, unsigned short len);

#endif //USB_H
